// SCCS:  SCCS: @(#) tcl16.rc 1.10 96/04/12 15:39:00
//
// Version
//

#include <patchlevel.h>

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 	7,5,1,0
 PRODUCTVERSION 7,5,1,0
 FILEFLAGSMASK 	0x3fL
 FILEFLAGS 	0x0L
 FILEOS 	0x1L
 FILETYPE	0x2L
 FILESUBTYPE 	0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "FileDescription", "Tcl16 DLL, 16-bit thunking module\0"
            VALUE "OriginalFilename", "tcl1675.dll\0"
            VALUE "CompanyName", "Sun Microsystems, Inc\0"
            VALUE "FileVersion", TCL_PATCH_LEVEL
            VALUE "LegalCopyright", "Copyright \251 1995-1996\0"
            VALUE "ProductName", "Tcl 7.5 for Windows\0"
            VALUE "ProductVersion", TCL_PATCH_LEVEL
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END


